import uuid
from typing import AsyncGenerator

from fastapi import APIRouter
from fastapi import Depends
from fastapi_users import BaseUserManager
from fastapi_users import FastAPIUsers
from fastapi_users import UUIDIDMixin
from fastapi_users.authentication import AuthenticationBackend
from fastapi_users.authentication import BearerTransport
from fastapi_users.authentication import CookieTransport
from fastapi_users.authentication import JWTStrategy
from fastapi_users_db_sqlmodel import SQLModelUserDatabaseAsync
from sqlalchemy.ext.asyncio import AsyncSession

from ...config import settings
from ..db import get_db
from ..models.security import OAuthAccount
from ..models.security import UserCreate
from ..models.security import UserOAuth as User
from ..models.security import UserRead
from ..models.security import UserUpdate


async def get_user_db(
    session: AsyncSession = Depends(get_db),
) -> SQLModelUserDatabaseAsync:
    yield SQLModelUserDatabaseAsync(session, User, OAuthAccount)


class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
    pass


async def get_user_manager(
    user_db: SQLModelUserDatabaseAsync = Depends(get_user_db),
) -> AsyncGenerator[UserManager, None]:
    yield UserManager(user_db)


bearer_transport = BearerTransport(tokenUrl="/auth/token/login")
cookie_transport = CookieTransport(cookie_samesite="none")


def get_jwt_strategy() -> JWTStrategy:
    return JWTStrategy(
        secret=settings.JWT_SECRET_KEY,
        lifetime_seconds=settings.JWT_EXPIRE_SECONDS,
    )


def get_jwt_cookie_strategy() -> JWTStrategy:
    return JWTStrategy(
        secret=settings.JWT_SECRET_KEY,
        lifetime_seconds=settings.COOKIE_EXPIRE_SECONDS,
    )


token_backend = AuthenticationBackend(
    name="bearer-jwt",
    transport=bearer_transport,
    get_strategy=get_jwt_strategy,
)
cookie_backend = AuthenticationBackend(
    name="cookie-jwt",
    transport=cookie_transport,
    get_strategy=get_jwt_cookie_strategy,
)


fastapi_users = FastAPIUsers[User, uuid.UUID](
    get_user_manager,
    [token_backend, cookie_backend],
)


current_active_user = fastapi_users.current_user(active=True)


# AUTH ROUTES

auth_router = APIRouter()

auth_router.include_router(
    fastapi_users.get_auth_router(token_backend),
    prefix="/token",
)
auth_router.include_router(
    fastapi_users.get_auth_router(cookie_backend),
)
auth_router.include_router(
    fastapi_users.get_register_router(UserRead, UserCreate),
)
auth_router.include_router(
    fastapi_users.get_reset_password_router(),
)
auth_router.include_router(
    fastapi_users.get_verify_router(UserRead),
)
auth_router.include_router(
    fastapi_users.get_users_router(UserRead, UserUpdate),
    prefix="/users",
)

# OAUTH CLIENTS
for client in settings.OAUTH_CLIENTS:
    # INIT CLIENTS
    client_name = client.CLIENT_NAME.lower()
    if client_name == "github":
        from httpx_oauth.clients.github import GitHubOAuth2

        _client = GitHubOAuth2(client.CLIENT_ID, client.CLIENT_SECRET)
    else:  # GENERIC CLIENT
        from httpx_oauth.oauth2 import OAuth2

        _client = OAuth2(
            client.CLIENT_ID,
            client.CLIENT_SECRET,
            client.AUTHORIZE_ENDPOINT,
            client.ACCESS_TOKEN_ENDPOINT,
            refresh_token_endpoint=client.REFRESH_TOKEN_ENDPOINT,
            revoke_token_endpoint=client.REVOKE_TOKEN_ENDPOINT,
        )

    # ADD ROUTES
    # GitHub OAuth
    auth_router.include_router(
        fastapi_users.get_oauth_router(
            _client,
            cookie_backend,
            settings.JWT_SECRET_KEY,
            # WARNING:
            # associate_by_email=True exposes to security risks if the OAuth
            # provider does not verify emails.
            associate_by_email=True,
        ),
        prefix=f"/{client_name}",
    )
    auth_router.include_router(
        fastapi_users.get_oauth_associate_router(
            _client, UserRead, settings.JWT_SECRET_KEY
        ),
        prefix=f"/{client_name}/associate",
    )



def collect_tasks():
    from fractal_tasks_core import __FRACTAL_MANIFEST__

    return (
        task_manifest
        for task_manifest in __FRACTAL_MANIFEST__
        if "task" in task_manifest["resource_type"]
    )



from typing import AsyncGenerator
from typing import Generator

from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import Session as DBSyncSession
from sqlalchemy.orm import sessionmaker

from ...config import settings

"""
Losely adapted from https://testdriven.io/blog/fastapi-sqlmodel/#async-sqlmodel
"""


engine = create_async_engine(
    settings.DATABASE_URL, echo=settings.DB_ECHO, future=True
)

engine_sync = create_engine(
    settings.DATABASE_SYNC_URL,
    echo=settings.DB_ECHO,
    future=True,
    connect_args={"check_same_thread": False},
)

async_session_maker = sessionmaker(
    engine, class_=AsyncSession, expire_on_commit=False
)


sync_session_maker = sessionmaker(
    bind=engine_sync, autocommit=False, autoflush=False
)


async def get_db() -> AsyncGenerator[AsyncSession, None]:
    async with async_session_maker() as async_session:
        yield async_session


def get_sync_db() -> Generator[DBSyncSession, None, None]:
    with sync_session_maker() as sync_session:
        yield sync_session



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

__VERSION__ = "0.1.5"


def collect_routers(app: FastAPI) -> None:
    from .app.api import router_default
    from .app.api import router_v1
    from .app.security import auth_router

    app.include_router(router_default, prefix="/api")
    app.include_router(router_v1, prefix="/api/v1")
    app.include_router(auth_router, prefix="/auth", tags=["auth"])


async def __on_startup():
    from .app.api.v1.task import collect_tasks_headless

    await collect_tasks_headless()


def start_application() -> FastAPI:
    app = FastAPI()
    collect_routers(app)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["http://127.0.0.1:5173", "http://localhost:5173"],
        allow_methods=["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"],
        allow_headers=[
            "set-cookie",
            "Set-Cookie",
            "Content-Type",
            "Access-Control-Allow-Headers",
            "X-Requested-With",
        ],
        allow_credentials=True,
    )

    return app



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
from . import __on_startup
from . import start_application


app = start_application()


@app.on_event("startup")
async def on_startup():
    await __on_startup()



import importlib
from concurrent.futures import Future
from copy import deepcopy
from logging import FileHandler
from logging import Formatter
from logging import getLogger
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from parsl.app.app import join_app
from parsl.app.python import PythonApp
from parsl.dataflow.dflow import DataFlowKernel
from parsl.dataflow.futures import AppFuture
from sqlalchemy.ext.asyncio import AsyncSession

from ... import __VERSION__
from ..models.project import Dataset
from ..models.project import Project
from ..models.task import PreprocessedTask
from ..models.task import Subtask
from ..models.task import Task
from .runner_utils import async_wrap
from .runner_utils import get_unique_executor
from .runner_utils import load_parsl_config


formatter = Formatter("%(asctime)s; %(levelname)s; %(message)s")
logger = getLogger(__name__)
handler = FileHandler("fractal.log", mode="a")
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel("INFO")


def _task_fun(
    *,
    task: Task,
    input_paths: List[Path],
    output_path: Path,
    metadata: Optional[Dict[str, Any]],
    task_args: Optional[Dict[str, Any]],
    executors: Union[
        List[str], Literal["all"]
    ],  # This is only needed for logging
    inputs,
):

    # NOTE: logging takes place here (in the function, not in the app), so that
    # it is only triggered when the function executes, rather than when the app
    # is defined. The executors argument is only needed for logging, in this
    # function.
    logger.info(f'Starting "{task.name}" task on {executors=}.')

    task_module = importlib.import_module(task.import_path)
    _callable = getattr(task_module, task.callable)
    metadata_update = _callable(
        input_paths=input_paths,
        output_path=output_path,
        metadata=metadata,
        **task_args,
    )
    metadata.update(metadata_update)
    try:
        metadata["history"].append(task.name)
    except KeyError:
        metadata["history"] = [task.name]
    return metadata


def _task_app_future(
    *,
    task: Task,
    input_paths: List[Path],
    output_path: Path,
    metadata: Optional[Dict[str, Any]],
    task_args: Optional[Dict[str, Any]],
    inputs,
    executors: Union[List[str], Literal["all"]] = "all",
    data_flow_kernel=None,
) -> AppFuture:

    app = PythonApp(
        _task_fun, executors=executors, data_flow_kernel=data_flow_kernel
    )
    # TODO: can we reassign app.__name__, for clarity in monitoring?
    return app(
        task=task,
        input_paths=input_paths,
        output_path=output_path,
        metadata=metadata,
        task_args=task_args,
        inputs=inputs,
        executors=executors,
    )


def _task_component_fun(
    *,
    task: Task,
    component: str,
    input_paths: List[Path],
    output_path: Path,
    metadata: Optional[Dict[str, Any]],
    task_args: Optional[Dict[str, Any]],
):

    logger.info(f'  Starting "{task.name}" for {component=}.')

    task_module = importlib.import_module(task.import_path)
    _callable = getattr(task_module, task.callable)
    _callable(
        input_paths=input_paths,
        output_path=output_path,
        metadata=metadata,
        component=component,
        **task_args,
    )
    return task.name, component


def _task_parallel_collect(
    metadata,
    task_name: str = None,
    component_list: Iterable[str] = None,
    inputs=None,
):
    history = f"{task_name}: {component_list}"
    try:
        metadata["history"].append(history)
    except KeyError:
        metadata["history"] = [history]

    return metadata


def _atomic_task_factory(
    *,
    task: Union[Task, Subtask, PreprocessedTask],
    input_paths: List[Path],
    output_path: Path,
    data_flow_kernel: DataFlowKernel,
    metadata: Optional[Union[Future, Dict[str, Any]]] = None,
    depends_on: Optional[List[AppFuture]] = None,
    workflow_id: int = None,
) -> AppFuture:
    """
    Single task processing

    Create a single PARSL app that encapsulates the task at hand and
    its parallelizazion.
    """
    if depends_on is None:
        depends_on = []

    task_args = task._arguments

    try:
        task_executor = get_unique_executor(
            workflow_id=workflow_id,
            task_executor=task.executor,
            data_flow_kernel=data_flow_kernel,
        )
    except ValueError as e:
        # When assigning a task to an unknown executor, make sure to cleanup
        # DFK before raising an error
        # data_flow_kernel.cleanup()
        logger.info("END of workflow due to ValueError (unknown executors).")
        raise ValueError(str(e))

    parall_level = task.parallelization_level
    if metadata and parall_level:

        # Define a single app
        task_component_app = PythonApp(
            _task_component_fun,
            executors=[task_executor],
            data_flow_kernel=data_flow_kernel,
        )

        # Define an app that takes all the other as input
        parallel_collection_app = PythonApp(
            _task_parallel_collect,
            executors="all",
            data_flow_kernel=data_flow_kernel,
        )

        @join_app(data_flow_kernel=data_flow_kernel)
        def _parallel_task_app_future(
            *,
            task: Task,
            parall_level: str,
            input_paths: List[Path],
            output_path: Path,
            metadata: AppFuture,
            task_args: Optional[Dict[str, Any]],
            executors: Union[List[str], Literal["all"]] = "all",
        ) -> AppFuture:

            logger.info(
                f'Starting {len(metadata[parall_level])} "{task.name}"'
                f" tasks in parallel, on {executors=}."
            )

            # Define a list of futures, to be used as inputs (AKA dependencies)
            # for parallel_collection_app. This must happen within a join_app,
            # because metadata has not yet been computed.
            app_futures = []
            for item in metadata[parall_level]:
                component_app_future = task_component_app(
                    task=task,
                    component=item,
                    input_paths=input_paths,
                    output_path=output_path,
                    metadata=metadata,
                    task_args=task_args,
                )
                app_futures.append(component_app_future)

            # Return the future for the app collecting all parallel tasks
            parallel_collection_app_future = parallel_collection_app(
                metadata,
                task_name=task.name,
                component_list=metadata[parall_level],
                inputs=app_futures,
            )

            return parallel_collection_app_future

        res = _parallel_task_app_future(
            parall_level=parall_level,
            task=task,
            input_paths=input_paths,
            output_path=output_path,
            metadata=metadata,
            task_args=task_args,
            executors=[task_executor],
        )
        return res
    else:
        res = _task_app_future(
            task=task,
            input_paths=input_paths,
            output_path=output_path,
            metadata=metadata,
            task_args=task_args,
            inputs=depends_on,
            executors=[task_executor],
            data_flow_kernel=data_flow_kernel,
        )
        return res


def _process_workflow(
    task: Union[Task, Subtask],
    input_paths: List[Path],
    output_path: Path,
    metadata: Dict[str, Any],
) -> AppFuture:
    """
    Creates the PARSL app that will execute the full workflow, taking care of
    dependencies

    Arguments
    ---------
    output_path (Path):
        directory or file where the final output, i.e., the output of the last
        task, will be written
    TBD

    Return
    ------
    TBD
    """
    preprocessed = task.preprocess()

    this_input = input_paths
    this_output = output_path
    this_metadata = deepcopy(metadata)

    workflow_id = task.id
    workflow_name = task.name
    dfk = load_parsl_config(
        workflow_id=workflow_id, workflow_name=workflow_name, logger=logger
    )

    app_futures: List[PythonApp] = []
    for i, task in enumerate(preprocessed):
        this_task_app_future = _atomic_task_factory(
            task=task,
            input_paths=this_input,
            output_path=this_output,
            metadata=app_futures[i - 1] if i > 0 else this_metadata,
            workflow_id=workflow_id,
            data_flow_kernel=dfk,
        )

        app_futures.append(this_task_app_future)
        this_input = [this_output]

    # Got to make sure that it is executed serially, task by task
    return app_futures[-1], dfk


async def auto_output_dataset(
    *,
    project: Project,
    input_dataset: Dataset,
    workflow: Task,
    overwrite_input: bool = False,
):
    """
    Determine the output dataset if it was not provided explicitly

    Only datasets containing exactly one path can be used as output.

    Returns
    -------
    output_dataset (Dataset):
        the output dataset
    """
    if overwrite_input and not input_dataset.read_only:
        input_paths = input_dataset.paths
        if len(input_paths) != 1:
            raise ValueError
        output_dataset = input_dataset
    else:
        raise NotImplementedError

    return output_dataset


def validate_workflow_compatibility(
    *,
    input_dataset: Dataset,
    workflow: Task,
    output_dataset: Optional[Dataset] = None,
):
    """
    Check compatibility of workflow and input / ouptut dataset
    """
    if (
        workflow.input_type != "Any"
        and workflow.input_type != input_dataset.type
    ):
        raise TypeError(
            f"Incompatible types `{workflow.input_type}` of workflow "
            f"`{workflow.name}` and `{input_dataset.type}` of dataset "
            f"`{input_dataset.name}`"
        )

    if not output_dataset:
        if input_dataset.read_only:
            raise ValueError("Input dataset is read-only")
        else:
            input_paths = input_dataset.paths
            if len(input_paths) != 1:
                # Only single input can be safely transformed in an output
                raise ValueError(
                    "Cannot determine output path: multiple input "
                    "paths to overwrite"
                )
            else:
                output_path = input_paths[0]
    else:
        output_path = output_dataset.paths
        if len(output_path) != 1:
            raise ValueError(
                "Cannot determine output path: Multiple paths in dataset."
            )
    return output_path


def get_app_future_result(app_future: AppFuture):
    """
    See issue #140 and https://stackoverflow.com/q/43241221/19085332

    By replacing
        .. = final_metadata.result()
    with
        .. = await async_wrap(get_app_future_result)(app_future=final_metadata)
    we avoid a (long) blocking statement.
    """
    return app_future.result()


async def submit_workflow(
    *,
    db: AsyncSession,
    workflow: Task,
    input_dataset: Dataset,
    output_dataset: Dataset,
):
    """
    Prepares a workflow and applies it to a dataset

    Arguments
    ---------
    db: (AsyncSession):
        Asynchronous database session
    output_dataset (Dataset | str) :
        the destination dataset of the workflow. If not provided, overwriting
        of the input dataset is implied and an error is raised if the dataset
        is in read only mode. If a string is passed and the dataset does not
        exist, a new dataset with that name is created and within it a new
        resource with the same name.
    """

    input_paths = input_dataset.paths
    output_path = output_dataset.paths[0]

    logger.info("*" * 80)
    logger.info(f"fractal_server.__VERSION__: {__VERSION__}")
    logger.info(f"START workflow {workflow.name}")
    logger.info(f"{input_paths=}")
    logger.info(f"{output_path=}")

    final_metadata, dfk = _process_workflow(
        task=workflow,
        input_paths=input_paths,
        output_path=output_path,
        metadata=input_dataset.meta,
    )
    output_dataset.meta = await async_wrap(get_app_future_result)(
        app_future=final_metadata
    )

    logger.info(f'END workflow "{workflow.name}"')

    # FIXME: This line is commented to avoid issue #94 of fractal-server
    # dfk.cleanup()

    db.add(output_dataset)

    await db.commit()



from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from fractal.common.models import SubtaskBase
from fractal.common.models import TaskBase
from pydantic import BaseModel
from sqlalchemy import Column
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.types import JSON
from sqlmodel import Field
from sqlmodel import Relationship

from ...utils import popget


def flatten(xx):
    """
    Flatten an arbitrarily nested sequence of sequences
    """
    for x in xx:
        if isinstance(x, PreprocessedTask):
            yield x
        else:
            yield from x


class PreprocessedTask(BaseModel):
    name: str
    module: str
    args: Dict[str, Any]
    save_intermediate_result: bool = False
    executor: Optional[str] = None
    parallelization_level: Optional[str] = None

    @property
    def callable(self):
        return self.module.partition(":")[2]

    @property
    def import_path(self):
        return self.module.partition(":")[0]

    @property
    def _arguments(self):
        return self.args


class Subtask(SubtaskBase, table=True):  # type: ignore
    class Config:
        arbitrary_types_allowed = True
        fields = {"parent": {"exclude": True}}

    id: Optional[int] = Field(default=None, primary_key=True)
    parent_task_id: Optional[int] = Field(
        default=None,
        foreign_key="task.id",
    )
    subtask_id: Optional[int] = Field(
        default=None,
        foreign_key="task.id",
    )
    parent: "Task" = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin="Subtask.parent_task_id==Task.id"
        )
    )
    subtask: "Task" = Relationship(
        sa_relationship_kwargs=dict(primaryjoin="Subtask.subtask_id==Task.id")
    )
    order: Optional[int]
    args: Dict[str, Any] = Field(sa_column=Column(JSON), default={})

    @property
    def _is_atomic(self):
        """
        A subtask is atomic iff the child task is atomic
        """
        return self.subtask._is_atomic

    @property
    def _arguments(self):
        """
        Override default arguments and strip specific arguments (executor and
        parallelization_level)
        """
        out = self.subtask.default_args.copy()
        out.update(self.args)
        popget(out, "executor")
        popget(out, "parallelization_level")
        return out

    @property
    def executor(self) -> Optional[str]:
        return self.args.get("executor") or self.subtask.executor

    @property
    def parallelization_level(self) -> Optional[str]:
        return (
            self.args.get("parallelization_level")
            or self.subtask.parallelization_level
        )

    @property
    def import_path(self):
        return self.subtask.import_path

    @property
    def callable(self):
        return self.subtask.callable

    def preprocess(self):
        if self._is_atomic:
            return PreprocessedTask(
                name=self.subtask.name,
                module=self.subtask.module,
                args=self._arguments,
                executor=self.executor,
                parallelization_level=self.parallelization_level,
            )
        else:
            return [st.preprocess() for st in self.subtask.subtask_list]

    @property
    def name(self):
        return self.subtask.name


class Task(TaskBase, table=True):  # type: ignore
    id: Optional[int] = Field(default=None, primary_key=True)
    default_args: Dict[str, Any] = Field(sa_column=Column(JSON), default={})
    subtask_list: List["Subtask"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin="Task.id==Subtask.parent_task_id",
            lazy="selectin",
            order_by="Subtask.order",
            collection_class=ordering_list("order"),
        ),
    )

    def preprocess(self):
        if not self._is_atomic:
            return flatten(st.preprocess() for st in self.subtask_list)
        else:
            return [
                PreprocessedTask(
                    name=self.name,
                    module=self.module,
                    args=self._arguments,
                    executor=self.executor,
                    parallelization_level=self.parallelization_level,
                )
            ]

    @property
    def _arguments(self):
        if not self._is_atomic:
            raise ValueError("Cannot call _arguments on a non-atomic task")
        out = self.default_args.copy()
        popget(out, "executor")
        popget(out, "parallelization_level")
        return out

    @property
    def executor(self) -> Optional[str]:
        try:
            return self.default_args["executor"]
        except KeyError:
            return None

    @property
    def parallelization_level(self) -> Optional[str]:
        try:
            return self.default_args["parallelization_level"]
        except KeyError:
            return None

    @property
    def module(self):
        return self.subtask.module

    @property
    def _is_atomic(self) -> bool:
        """
        A task is atomic iff it does not contain subtasks
        """
        return not self.subtask_list

    @property
    def callable(self):
        return self.module.partition(":")[2]

    @property
    def import_path(self):
        return self.module.partition(":")[0]

    async def add_subtask(
        self,
        db: AsyncSession,
        subtask_id: Optional[int] = None,
        subtask: Optional["Task"] = None,
        order: Optional[int] = None,
        args: Optional[Dict[str, Any]] = None,
        commit_and_refresh: bool = True,
    ):
        if subtask is None:
            subtask = await db.get(Task, subtask_id)

        if not args:
            args = dict()
        st = Subtask(parent=self, subtask=subtask, args=args)

        if order is None:
            self.subtask_list.append(st)
        else:
            self.subtask_list.insert(order, st)

        db.add_all([self, st])
        if commit_and_refresh:
            await db.commit()
            await db.refresh(self)



"""empty message

Revision ID: 7bb51084235d
Revises: 
Create Date: 2022-09-08 15:08:47.669322

"""
from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlmodel


# revision identifiers, used by Alembic.
revision = '7bb51084235d'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('task',
    sa.Column('default_args', sa.JSON(), nullable=True),
    sa.Column('name', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('resource_type', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('module', sqlmodel.sql.sqltypes.AutoString(), nullable=True),
    sa.Column('input_type', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('output_type', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )
    op.create_table('user_oauth',
    sa.Column('id', sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
    sa.Column('email', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('hashed_password', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('is_active', sa.Boolean(), nullable=False),
    sa.Column('is_superuser', sa.Boolean(), nullable=False),
    sa.Column('is_verified', sa.Boolean(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_user_oauth_email'), 'user_oauth', ['email'], unique=True)
    op.create_table('oauthaccount',
    sa.Column('id', sqlmodel.sql.sqltypes.GUID(), nullable=False),
    sa.Column('user_id', sqlmodel.sql.sqltypes.GUID(), nullable=False),
    sa.Column('oauth_name', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('access_token', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('expires_at', sa.Integer(), nullable=True),
    sa.Column('refresh_token', sqlmodel.sql.sqltypes.AutoString(), nullable=True),
    sa.Column('account_id', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('account_email', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.ForeignKeyConstraint(['user_id'], ['user_oauth.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_oauthaccount_account_id'), 'oauthaccount', ['account_id'], unique=False)
    op.create_index(op.f('ix_oauthaccount_oauth_name'), 'oauthaccount', ['oauth_name'], unique=False)
    op.create_table('project',
    sa.Column('name', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('project_dir', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('read_only', sa.Boolean(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('slug', sqlmodel.sql.sqltypes.AutoString(), nullable=True),
    sa.Column('user_owner_id', sqlmodel.sql.sqltypes.GUID(), nullable=True),
    sa.ForeignKeyConstraint(['user_owner_id'], ['user_oauth.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('subtask',
    sa.Column('args', sa.JSON(), nullable=True),
    sa.Column('parent_task_id', sa.Integer(), nullable=False),
    sa.Column('subtask_id', sa.Integer(), nullable=False),
    sa.Column('order', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['parent_task_id'], ['task.id'], ),
    sa.ForeignKeyConstraint(['subtask_id'], ['task.id'], ),
    sa.PrimaryKeyConstraint('parent_task_id', 'subtask_id')
    )
    op.create_table('dataset',
    sa.Column('meta', sa.JSON(), nullable=True),
    sa.Column('name', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('type', sqlmodel.sql.sqltypes.AutoString(), nullable=True),
    sa.Column('read_only', sa.Boolean(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['project_id'], ['project.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('resource',
    sa.Column('path', sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    sa.Column('glob_pattern', sqlmodel.sql.sqltypes.AutoString(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('dataset_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['dataset_id'], ['dataset.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('resource')
    op.drop_table('dataset')
    op.drop_table('subtask')
    op.drop_table('project')
    op.drop_index(op.f('ix_oauthaccount_oauth_name'), table_name='oauthaccount')
    op.drop_index(op.f('ix_oauthaccount_account_id'), table_name='oauthaccount')
    op.drop_table('oauthaccount')
    op.drop_index(op.f('ix_user_oauth_email'), table_name='user_oauth')
    op.drop_table('user_oauth')
    op.drop_table('task')
    # ### end Alembic commands ###



import uuid
from typing import List
from typing import Optional

from fastapi_users import schemas
from fastapi_users_db_sqlmodel import SQLModelBaseOAuthAccount
from fastapi_users_db_sqlmodel import SQLModelBaseUserDB
from pydantic import UUID4
from sqlalchemy_utils import UUIDType
from sqlmodel import Column
from sqlmodel import Field
from sqlmodel import Relationship


"""
Adapted from
    https://github.com/fastapi-users/fastapi-users-db-sqlmodel/
        blob/main/tests/conftest.py
"""


class UserOAuth(SQLModelBaseUserDB, table=True):
    __tablename__ = "user_oauth"
    id: UUID4 = Field(
        default_factory=uuid.uuid4,
        nullable=False,
        sa_column=Column(UUIDType(), primary_key=True),
    )
    oauth_accounts: List["OAuthAccount"] = Relationship(
        back_populates="user",
        sa_relationship_kwargs={"lazy": "selectin", "cascade": "all, delete"},
    )


class OAuthAccount(SQLModelBaseOAuthAccount, table=True):
    user_id: UUID4 = Field(foreign_key="user_oauth.id", nullable=False)
    user: Optional[UserOAuth] = Relationship(back_populates="oauth_accounts")


class UserRead(schemas.BaseUser[uuid.UUID]):
    pass


class UserUpdate(schemas.BaseUserUpdate):
    pass


class UserCreate(schemas.BaseUserCreate):
    pass



import logging
from typing import List

from fastapi import APIRouter
from fastapi import BackgroundTasks
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Response
from fastapi import status
from sqlalchemy.exc import IntegrityError
from sqlmodel import select

from ...db import AsyncSession
from ...db import DBSyncSession
from ...db import get_db
from ...db import get_sync_db
from ...models import ApplyWorkflow
from ...models import Dataset
from ...models import DatasetCreate
from ...models import DatasetRead
from ...models import DatasetUpdate
from ...models import Project
from ...models import ProjectCreate
from ...models import ProjectRead
from ...models import Resource
from ...models import ResourceCreate
from ...models import ResourceRead
from ...models.task import Task
from ...runner import auto_output_dataset
from ...runner import submit_workflow
from ...runner import validate_workflow_compatibility
from ...security import current_active_user
from ...security import User

router = APIRouter()


@router.get("/", response_model=List[ProjectRead])
async def get_list_project(
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    stm = select(Project).where(Project.user_owner_id == user.id)
    res = await db.execute(stm)
    project_list = res.scalars().all()
    return project_list


@router.get("/{project_id}", response_model=ProjectRead)
async def get_project(
    project_id: int,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    project = await db.get(Project, project_id)
    if not project:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Project not found"
        )
    if project.user_owner_id != user.id:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Not allowed on project",
        )
    return project


@router.delete("/{project_id}", status_code=204)
async def delete_project(
    project_id: int,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    project = await db.get(Project, project_id)
    if not project:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Project not found"
        )
    if project.user_owner_id != user.id:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Not allowed on project",
        )
    await db.delete(project)
    await db.commit()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post("/", response_model=ProjectRead, status_code=201)
async def create_project(
    project: ProjectCreate,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    """
    Create new poject
    """

    # Check that there is no project with the same user and name
    stm = (
        select(Project)
        .where(Project.user_owner_id == user.id)
        .where(Project.name == project.name)
    )
    res = await db.execute(stm)
    if res.scalars().all():
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Project name ({project.name}) already in use",
        )

    db_project = Project.from_orm(project)
    db_project.dataset_list.append(Dataset(name=project.default_dataset_name))

    db_project.user_owner_id = user.id
    try:
        db.add(db_project)
        await db.commit()
        await db.refresh(db_project)
    except IntegrityError as e:
        await db.rollback()
        logging.error(str(e))
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=str(e),
        )

    return db_project


@router.post(
    "/apply/",
    status_code=status.HTTP_202_ACCEPTED,
)
async def apply_workflow(
    apply_workflow: ApplyWorkflow,
    background_tasks: BackgroundTasks,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
    db_sync: DBSyncSession = Depends(get_sync_db),
):
    stm = (
        select(Project, Dataset)
        .join(Dataset)
        .where(Project.user_owner_id == user.id)
        .where(Project.id == apply_workflow.project_id)
        .where(Dataset.id == apply_workflow.input_dataset_id)
    )
    project, input_dataset = (await db.execute(stm)).one()

    # TODO check that user is allowed to use this task

    workflow = db_sync.get(Task, apply_workflow.workflow_id)

    if apply_workflow.output_dataset_id:
        stm = (
            select(Dataset)
            .where(Dataset.project_id == project.id)
            .where(Dataset.id == apply_workflow.output_dataset_id)
        )
        output_dataset = (await db.execute(stm)).scalars().one()
    else:
        output_dataset = await auto_output_dataset(
            project=project, input_dataset=input_dataset, workflow=workflow
        )

    try:
        validate_workflow_compatibility(
            workflow=workflow,
            input_dataset=input_dataset,
            output_dataset=output_dataset,
        )
    except TypeError as e:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e)
        )

    if not input_dataset or not output_dataset or not workflow:
        raise ValueError

    background_tasks.add_task(
        submit_workflow,
        workflow=workflow,
        input_dataset=input_dataset,
        output_dataset=output_dataset,
        db=db,
    )

    # TODO we should return a job id of some sort
    return dict(status="submitted")


@router.post(
    "/{project_id}/",
    response_model=DatasetRead,
    status_code=status.HTTP_201_CREATED,
)
async def add_dataset(
    project_id: int,
    dataset: DatasetCreate,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    """
    Add new dataset to current project
    """
    project = await db.get(Project, project_id)
    if project.user_owner_id != user.id:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Not allowed on project",
        )
    dataset.project_id = project.id
    db_dataset = Dataset.from_orm(dataset)
    db.add(db_dataset)
    await db.commit()
    await db.refresh(db_dataset)
    return db_dataset


@router.get(
    "/{project_id}/{dataset_id}",
    response_model=List[ResourceRead],
)
async def get_resource(
    project_id: int,
    dataset_id: int,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    """
    Get resources from a dataset
    """
    project = await db.get(Project, project_id)
    if project.user_owner_id != user.id:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Not allowed on project",
        )

    stm = select(Resource).where(Resource.dataset_id == dataset_id)
    res = await db.execute(stm)
    resource_list = res.scalars().all()
    return resource_list


@router.delete("/{project_id}/{dataset_id}/{resource_id}", status_code=204)
async def delete_resource(
    project_id: int,
    dataset_id: int,
    resource_id: int,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    project = await db.get(Project, project_id)
    if not project:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Project not found"
        )
    if project.user_owner_id != user.id:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Not allowed on project",
        )
    resource = await db.get(Resource, resource_id)
    if not resource or resource.dataset_id not in (
        ds.id for ds in project.dataset_list
    ):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Resource does not exist or does not belong to project",
        )
    await db.delete(resource)
    await db.commit()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post(
    "/{project_id}/{dataset_id}",
    response_model=ResourceRead,
    status_code=status.HTTP_201_CREATED,
)
async def add_resource(
    project_id: int,
    dataset_id: int,
    resource: ResourceCreate,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    """
    Add resource to an existing dataset
    """
    stm = (
        select(Project, Dataset)
        .join(Dataset)
        .where(Project.user_owner_id == user.id)
        .where(Project.id == project_id)
        .where(Dataset.id == dataset_id)
    )
    project, dataset = (await db.execute(stm)).one()
    db_resource = Resource(dataset_id=dataset.id, **resource.dict())
    db.add(db_resource)
    await db.commit()
    await db.refresh(db_resource)
    return db_resource


@router.patch("/{project_id}/{dataset_id}", response_model=DatasetRead)
async def patch_dataset(
    project_id: int,
    dataset_id: int,
    dataset_update: DatasetUpdate,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):

    project = await db.get(Project, project_id)
    if project.user_owner_id != user.id:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Not allowed on project",
        )

    db_dataset = await db.get(Dataset, dataset_id)
    for key, value in dataset_update.dict(exclude_unset=True).items():
        setattr(db_dataset, key, value)

    await db.commit()
    await db.refresh(db_dataset)
    return db_dataset



"""empty message

Revision ID: 4aa8e76fb21d
Revises: 7bb51084235d
Create Date: 2022-09-08 15:16:32.855071

"""
from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlmodel


# revision identifiers, used by Alembic.
revision = '4aa8e76fb21d'
down_revision = '7bb51084235d'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from fastapi import APIRouter
from fastapi import Depends
from sqlmodel import select

from ...db import AsyncSession
from ...db import get_db
from ...models import Dataset
from ...models import DatasetRead
from ...models import Project
from ...security import current_active_user
from ...security import User


router = APIRouter()


@router.get("/{project_id}/{dataset_id}", response_model=DatasetRead)
async def get_dataset(
    project_id: int,
    dataset_id: int,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    stm = (
        select(Project, Dataset)
        .join(Dataset)
        .where(Project.user_owner_id == user.id)
        .where(Project.id == project_id)
        .where(Dataset.id == dataset_id)
    )
    project, dataset = (await db.execute(stm)).one()
    return dataset



from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from fractal.common.models import DatasetBase
from fractal.common.models import ProjectBase
from fractal.common.models import ResourceBase
from pydantic import UUID4
from sqlalchemy import Column
from sqlalchemy.types import JSON
from sqlmodel import Field
from sqlmodel import Relationship

from .security import UserOAuth


class Dataset(DatasetBase, table=True):  # type: ignore
    id: Optional[int] = Field(default=None, primary_key=True)
    project_id: int = Field(foreign_key="project.id")
    resource_list: List["Resource"] = Relationship(
        sa_relationship_kwargs={"lazy": "selectin"}
    )
    meta: Dict[str, Any] = Field(sa_column=Column(JSON), default={})

    class Config:
        arbitrary_types_allowed = True

    @property
    def paths(self) -> List[Path]:
        return [r.glob_path for r in self.resource_list]


class Project(ProjectBase, table=True):  # type: ignore
    id: Optional[int] = Field(default=None, primary_key=True)
    slug: Optional[str] = Field()

    user_owner_id: Optional[UUID4] = Field(foreign_key="user_oauth.id")
    user: Optional[UserOAuth] = Relationship()

    dataset_list: List[Dataset] = Relationship(
        sa_relationship_kwargs={
            "lazy": "selectin",
            "cascade": "all, delete-orphan",
        }
    )


class Resource(ResourceBase, table=True):  # type: ignore
    id: Optional[int] = Field(default=None, primary_key=True)
    dataset_id: int = Field(foreign_key="dataset.id")



from fastapi import APIRouter

from ...config import settings
from .v1.dataset import router as dataset_router
from .v1.project import router as project_router
from .v1.task import router as task_router


router_default = APIRouter()
router_v1 = APIRouter()

router_v1.include_router(project_router, prefix="/project", tags=["project"])
router_v1.include_router(dataset_router, prefix="/dataset", tags=["dataset"])
router_v1.include_router(task_router, prefix="/task", tags=["task"])
router_v1.include_router(task_router, prefix="/workflow", tags=["workflow"])


@router_default.get("/alive/")
async def alive():
    return dict(
        alive=True,
        deployment_type=settings.DEPLOYMENT_TYPE,
        version=settings.PROJECT_VERSION,
    )



from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Security
from fastapi import status
from fastapi.security import OAuth2PasswordRequestForm

from ...security import FailedAuthenticationException
from ...security import get_current_user
from ...security import Token
from ...security import token_authentication_login
from ...security import User


router = APIRouter()


@router.post("/token", response_model=Token)
async def login_for_access_token(
    form_data: OAuth2PasswordRequestForm = Depends(),
):
    try:
        return await token_authentication_login(
            form_data.username, form_data.password
        )
    except FailedAuthenticationException:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
        )


@router.get("/me", response_model=User)
async def who_am_i(
    current_user: User = Security(get_current_user, scopes=["profile"])
):
    return current_user



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Marco Franzon <marco.franzon@exact-lab.it>
Tommaso Comparin <tommaso.comparin@exact-lab.it>
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>


This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import asyncio
import logging
from functools import partial
from functools import wraps
from typing import Callable

import parsl
from parsl.addresses import address_by_hostname
from parsl.channels import LocalChannel
from parsl.config import Config
from parsl.dataflow.dflow import DataFlowKernelLoader
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SingleNodeLauncher
from parsl.launchers import SrunLauncher
from parsl.monitoring.monitoring import MonitoringHub
from parsl.providers import LocalProvider
from parsl.providers import SlurmProvider

from ...config import settings


def add_prefix(*, workflow_id: int, executor_label: str):
    return f"{workflow_id}___{executor_label}"


def async_wrap(func: Callable) -> Callable:
    """
    See issue #140 and https://stackoverflow.com/q/43241221/19085332

    By replacing
        .. = final_metadata.result()
    with
        .. = await async_wrap(get_app_future_result)(app_future=final_metadata)
    we avoid a (long) blocking statement.
    """

    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)

    return run


def generate_parsl_config(
    *,
    workflow_id: int,
    workflow_name: str,
    enable_monitoring: bool = True,
) -> Config:
    allowed_configs = ["local", "pelkmanslab", "fmi", "custom"]
    config = settings.PARSL_CONFIG
    if config not in allowed_configs:
        raise ValueError(f"{config=} not in {allowed_configs=}")
    if config == "custom":
        raise NotImplementedError

    if config == "local":
        # Define a single provider
        prov_local = LocalProvider(
            launcher=SingleNodeLauncher(debug=False),
            channel=LocalChannel(),
            init_blocks=1,
            min_blocks=0,
            max_blocks=4,
        )
        # Define executors
        providers = [prov_local] * 4
        labels = ["cpu-low", "cpu-mid", "cpu-high", "gpu"]
        executors = []
        for provider, label in zip(providers, labels):
            executors.append(
                HighThroughputExecutor(
                    label=add_prefix(
                        workflow_id=workflow_id, executor_label=label
                    ),
                    provider=provider,
                    address=address_by_hostname(),
                    cpu_affinity="block",
                )
            )

    elif config == "pelkmanslab":

        # Define providers
        common_args = dict(
            partition="main",
            nodes_per_block=1,
            init_blocks=1,
            min_blocks=0,
            max_blocks=100,
            parallelism=1,
            exclusive=False,
            walltime="20:00:00",
        )
        prov_cpu_low = SlurmProvider(
            launcher=SrunLauncher(debug=False),
            channel=LocalChannel(),
            cores_per_node=1,
            mem_per_node=7,
            **common_args,
        )
        prov_cpu_mid = SlurmProvider(
            launcher=SrunLauncher(debug=False),
            channel=LocalChannel(),
            cores_per_node=4,
            mem_per_node=15,
            **common_args,
        )
        prov_cpu_high = SlurmProvider(
            launcher=SrunLauncher(debug=False),
            channel=LocalChannel(),
            cores_per_node=16,
            mem_per_node=61,
            **common_args,
        )
        # Define a gpu provider
        prov_gpu = SlurmProvider(
            partition="gpu",
            launcher=SrunLauncher(debug=False),
            channel=LocalChannel(),
            nodes_per_block=1,
            init_blocks=1,
            min_blocks=0,
            max_blocks=2,
            mem_per_node=61,
            walltime="10:00:00",
        )

        # Define executors
        providers = [prov_cpu_low, prov_cpu_mid, prov_cpu_high, prov_gpu]
        labels = ["cpu-low", "cpu-mid", "cpu-high", "gpu"]
        # FIXME
        list_mem_per_worker = [7, 15, 61, 61]  # FIXME
        executors = []
        for provider, label in zip(providers, labels):
            executors.append(
                HighThroughputExecutor(
                    label=add_prefix(
                        workflow_id=workflow_id, executor_label=label
                    ),
                    provider=provider,
                    mem_per_worker=list_mem_per_worker[
                        labels.index(label)
                    ],  # FIXME
                    max_workers=100,
                    address=address_by_hostname(),
                    cpu_affinity="block",
                )
            )

    elif config == "fmi":

        # FIXME: fix partition

        # Define a cpu provider
        provider_args = dict(
            partition="main",
            launcher=SrunLauncher(debug=False),
            channel=LocalChannel(),
            nodes_per_block=1,
            init_blocks=1,
            min_blocks=0,
            max_blocks=4,
            walltime="10:00:00",
            exclusive=False,
        )
        prov_slurm_cpu = SlurmProvider(**provider_args)

        # Define executors
        htex_slurm_cpu = HighThroughputExecutor(
            label=add_prefix(
                workflow_id=workflow_id, executor_label="cpu-low"
            ),
            provider=prov_slurm_cpu,
            address=address_by_hostname(),
            cpu_affinity="block",
        )
        htex_slurm_cpu_2 = HighThroughputExecutor(
            label=add_prefix(workflow_id=workflow_id, executor_label="cpu-2"),
            provider=prov_slurm_cpu,
            address=address_by_hostname(),
            cpu_affinity="block",
        )

        executors = [htex_slurm_cpu, htex_slurm_cpu_2]

    # Extract the executor labels
    # new_executor_labels = [executor.label for executor in executors]

    # Define monitoring hub and finalize configuration
    if enable_monitoring:
        monitoring = MonitoringHub(
            hub_address=address_by_hostname(),
            workflow_name=workflow_name,
        )
    else:
        monitoring = None

    config = Config(
        executors=executors, monitoring=monitoring, max_idletime=20.0
    )
    return config


def load_parsl_config(
    *,
    workflow_id: int,
    workflow_name: str = "default workflow name",
    config: Config = None,
    logger: logging.Logger = None,
    enable_monitoring: bool = True,
):

    if logger is None:
        logger = logging.getLogger("logs")

    logger.info(f"{settings.PARSL_CONFIG=}")
    logger.info(f"{settings.PARSL_DEFAULT_EXECUTOR=}")

    if not config:
        config = generate_parsl_config(
            enable_monitoring=enable_monitoring,
            workflow_id=workflow_id,
            workflow_name=workflow_name,
        )

    try:
        dfk = DataFlowKernelLoader.dfk()
        old_executor_labels = [
            executor_label for executor_label in dfk.executors.keys()
        ]
        logger.info(
            f"DFK {dfk} exists, with {len(dfk.executors)} executors: "
            f"{old_executor_labels}"
        )

        # FIXME: what if an executor was already there?
        # (re-submitting same workflow?)
        dfk.add_executors(config.executors)

    # FIXME: better exception handling
    except RuntimeError:
        logger.info(
            "DFK probably missing, "
            "proceed with parsl.clear and parsl.config.Config"
        )
        parsl.clear()
        DataFlowKernelLoader.load(config)
        dfk = DataFlowKernelLoader.dfk()

    executor_labels = [
        executor_label for executor_label in dfk.executors.keys()
    ]
    logger.info(
        f"DFK {dfk} now has {len(executor_labels)} executors: "
        f"{executor_labels}"
    )

    return dfk


def get_unique_executor(
    *, workflow_id: int, task_executor: str = None, data_flow_kernel=None
):

    # Handle missing value
    if task_executor is None:
        task_executor = settings.PARSL_DEFAULT_EXECUTOR

    # Redefine task_executor, by prepending workflow_id
    new_task_executor = add_prefix(
        workflow_id=workflow_id, executor_label=task_executor
    )

    # Verify match between new_task_executor and available executors
    valid_executor_labels = data_flow_kernel.executors.keys()
    if new_task_executor not in valid_executor_labels:
        raise ValueError(
            f"Executor label {new_task_executor} is not in "
            f"{valid_executor_labels=}"
        )

    return new_task_executor



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
from enum import Enum
from os import environ
from os import getenv
from os.path import abspath
from typing import List
from typing import Optional

from dotenv import load_dotenv
from pydantic import BaseModel
from pydantic import BaseSettings
from pydantic import Field
from pydantic import root_validator

import fractal_server


def fail_getenv(key):
    value = getenv(key, None)
    if value is None:
        raise ValueError(f"Must provide environment variable {key}")
    return value


load_dotenv(".fractal_server.env")


class OAuthClient(BaseModel):
    CLIENT_NAME: str
    CLIENT_ID: str
    CLIENT_SECRET: str

    AUTHORIZE_ENDPOINT: Optional[str]
    ACCESS_TOKEN_ENDPOINT: Optional[str]
    REFRESH_TOKEN_ENDPOINT: Optional[str]
    REVOKE_TOKEN_ENDPOINT: Optional[str]


class DeploymentType(str, Enum):
    PRODUCTION = "production"
    STAGING = "staging"
    TESTING = "testing"
    DEVELOPMENT = "development"


class Settings(BaseSettings):
    PROJECT_NAME: str = "Fractal Server"
    PROJECT_VERSION: str = fractal_server.__VERSION__
    DEPLOYMENT_TYPE: DeploymentType = DeploymentType(
        fail_getenv("DEPLOYMENT_TYPE")
    )

    ###########################################################################
    # AUTH
    ###########################################################################

    OAUTH_CLIENTS: List[OAuthClient] = Field(default_factory=list)

    # JWT TOKEN
    JWT_EXPIRE_SECONDS: int = int(getenv("JWT_EXPIRE_SECONDS", default=180))
    JWT_SECRET_KEY: str = fail_getenv("JWT_SECRET_KEY")

    # COOKIE TOKEN
    COOKIE_EXPIRE_SECONDS: int = int(
        getenv("COOKIE_EXPIRE_SECONDS", default=86400)
    )

    ###########################################################################
    # DATABASE
    ###########################################################################
    DB_ENGINE: str = getenv("DB_ENGINE", "sqlite")

    if DB_ENGINE == "postgres":
        POSTGRES_USER: str = fail_getenv("POSTGRES_USER")
        POSTGRES_PASSWORD: str = fail_getenv("POSTGRES_PASSWORD", "password")
        POSTGRES_SERVER: str = getenv("POSTGRES_SERVER", "localhost")
        POSTGRES_PORT: str = getenv("POSTGRES_PORT", "5432")
        POSTGRES_DB: str = fail_getenv("POSTGRES_DB")

        DATABASE_URL = (
            f"postgresql+asyncpg://{POSTGRES_USER}:{POSTGRES_PASSWORD}"
            f"@{POSTGRES_SERVER}:{POSTGRES_PORT}/{POSTGRES_DB}"
        )
        DATABASE_SYNC_URL = DATABASE_URL.replace("asyncpg", "psycopg2")
    elif DB_ENGINE == "sqlite":
        SQLITE_PATH: str = fail_getenv("SQLITE_PATH")

        DATABASE_URL = (
            "sqlite+aiosqlite:///"
            f"{abspath(SQLITE_PATH) if SQLITE_PATH else SQLITE_PATH}"
        )
        DATABASE_SYNC_URL = DATABASE_URL.replace("aiosqlite", "pysqlite")

    @property
    def DB_ECHO(self):
        db_echo = bool(
            int(
                getenv(
                    "DB_ECHO",
                    self.DEPLOYMENT_TYPE != DeploymentType.PRODUCTION,
                )
            )
        )
        return db_echo

    ###########################################################################
    # FRACTAL SPECIFIC
    ###########################################################################
    DATA_DIR_ROOT: str = fail_getenv("DATA_DIR_ROOT")
    PARSL_CONFIG: str = getenv("PARSL_CONFIG", "local")
    PARSL_DEFAULT_EXECUTOR: str = getenv("PARSL_DEFAULT_EXECUTOR", "cpu-low")

    @root_validator(pre=True)
    def collect_oauth_clients(cls, values):
        oauth_env_variable_keys = [
            key for key in environ.keys() if "OAUTH" in key
        ]
        clients_available = {
            var.split("_")[1] for var in oauth_env_variable_keys
        }

        values["OAUTH_CLIENTS"] = []
        for client in clients_available:
            prefix = f"OAUTH_{client}"
            oauth_client = OAuthClient(
                CLIENT_NAME=client,
                CLIENT_ID=getenv(f"{prefix}_CLIENT_ID", None),
                CLIENT_SECRET=getenv(f"{prefix}_CLIENT_SECRET", None),
                AUTHORIZE_ENDPOINT=getenv(
                    f"{prefix}_AUTHORIZE_ENDPOINT", None
                ),
                ACCESS_TOKEN_ENDPOINT=getenv(
                    f"{prefix}_ACCESS_TOKEN_ENDPOINT", None
                ),
                REFRESH_TOKEN_ENDPOINT=getenv(
                    f"{prefix}_REFRESH_TOKEN_ENDPOINT", None
                ),
                REVOKE_TOKEN_ENDPOINT=getenv(
                    f"{prefix}_REVOKE_TOKEN_ENDPOINT", None
                ),
            )
            values["OAUTH_CLIENTS"].append(oauth_client)
        return values


settings = Settings()



import argparse as ap
from sys import argv

import uvicorn

parser = ap.ArgumentParser()
parser.add_argument("--host", default="127.0.0.1")
parser.add_argument("-p", "--port", default=8000, type=int)
parser.add_argument("--reload", default=False, action="store_true")


def run():
    args = parser.parse_args(argv[1:])
    uvicorn.run(
        "fractal_server.main:app",
        host=args.host,
        port=args.port,
        reload=args.reload,
    )


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fractal_server',
 'fractal_server.app',
 'fractal_server.app.api',
 'fractal_server.app.api.v1',
 'fractal_server.app.db',
 'fractal_server.app.models',
 'fractal_server.app.runner',
 'fractal_server.app.security',
 'fractal_server.migrations',
 'fractal_server.migrations.versions']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy-Utils>=0.38.3,<0.39.0',
 'aiosqlite>=0.17.0,<0.18.0',
 'alembic>=1.8.0,<2.0.0',
 'fastapi-users-db-sqlmodel>=0.2.0,<0.3.0',
 'fastapi-users[oauth]>=10.1.1,<11.0.0',
 'fastapi>=0.78.0,<0.79.0',
 'fractal-client<1.0.0',
 'parsl>=1.3.0dev0,<2.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'sqlmodel>=0.0.8,<0.0.9',
 'uvicorn>=0.18.2,<0.19.0']

extras_require = \
{'tasks-core': ['fractal-tasks-core<1.0.0']}

entry_points = \
{'console_scripts': ['fractal-server = fractal_server.__main__:run']}

setup_kwargs = {
    'name': 'fractal-server',
    'version': '0.1.5',
    'description': 'Server component of the Fractal analytics platform',
    'long_description': '# Fractal Server\n\n[![PyPI version](https://img.shields.io/pypi/v/fractal-server?color=gree)](https://pypi.org/project/fractal-server/)\n\nFractal is a framework to process high content imaging data at scale and prepare it for interactive visualization.\n\nThis is the server component of the fractal analytics platform. If you are interested in the client component, please refer to the [main\nrepository](https://github.com/fractal-analytics-platform/fractal). If you are interested in the fractal tasks, please refer to [the tasks repository](https://github.com/fractal-analytics-platform/fractal-tasks-core).\n\n## Installation\n\nYou may\n`pip install fractal-server`\n\nThis will install the project and its dependencies. If you wish to also install\nFractal core tasks, use\n```\npip install fractal-server[core-tasks]\n```\n\n### Environment and database\n\nYou will need to define some environment variables in order to use\n`fractal-server`. For your convenience, you may simply copy\n`template.fractal_server.env` to `.fractal_server.env`.\n\n`fractal-server` requires a database to run. Once you set up the environment\nvariables you need to initialise the database by invoking\n\n```\nalembic upgrade head\n```\n\nNOTE: as `fractal-server` is still in pre-alpha the database schema is not yet\ncommitted to the repository. As such you\'ll be required to first create a\nschema revision with `alembic revision --autogenerate -m "[your commit\nmessage]"`.\n\n## Contributing\n\nTo contribute to the development of `fractal-server` you may fork and clone the\n[repository](https://github.com/fractal-analytics-platform/fractal-server).\n\nWe use [poetry](https://python-poetry.org/docs/) (v1.2) to manage the\ndevelopment environment and the dependencies. Running\n\n```\npoetry install [--with dev]\n```\n\nwill take care of installing all the dependencies in a separate environment,\noptionally installing also the development dependencies.\n\nIt may be useful to use a different interpreter from the one installed in your\nsystem. To this end we recommend using\n[pyenv](https://github.com/pyenv/pyenv). In the project folder, invoking\n\n```\npyenv local 3.8.13\npoetry env use 3.8\npoetry install\n```\n\nwill install Fractal in a development environment using `python 3.8.13` instad\nof the system-wide interpreter.\n\n### Testing\n\nWe use [pytest](https://docs.pytest.org/en/7.1.x/) for unit and integration\ntesting of Fractal. If you installed the development dependencies, you may run\nthe test suite by invoking\n\n```\npoetry run pytest\n```\n\n## Contributors\n\nFractal was conceived in the Liberali Lab at the Friedrich Miescher Institute\nfor Biomedical Research and in the Pelkmans Lab at the University of Zurich\n(both in Switzerland). The project lead is with\n[@gusqgm](https://github.com/gusqgm) & [@jluethi](https://github.com/jluethi).\nThe core development is done under contract by\n[@mfranzon](https://github.com/mfranzon), [@tcompa](https://github.com/tcompa)\n& [@jacopo-exact](https://github.com/jacopo-exact) from [eXact lab\nS.r.l.](https://exact-lab.it).\n\n## License\n\nFractal is released according to a BSD 3-Clause License. See `LICENSE`.\n',
    'author': 'Jacopo Nespolo',
    'author_email': 'jacopo.nespolo@exact-lab.it',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/fractal-analytics-platform/fractal-server',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
from collections.abc import MutableMapping
from typing import Any
from warnings import warn as _warn

from .config import DeploymentType
from .config import settings


def warn(message):
    """
    Make sure that warnings do not make their way to staing and production
    """
    if settings.DEPLOYMENT_TYPE in [
        DeploymentType.TESTING,
        DeploymentType.DEVELOPMENT,
    ]:
        _warn(message, RuntimeWarning)
    else:
        raise RuntimeError(message)


def popget(d: MutableMapping, key: str, default: Any = None) -> Any:
    """
    Pop and return mapping item if possible or return default
    """
    try:
        return d.pop(key)
    except KeyError:
        return default



import asyncio
from collections import Counter
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import status
from sqlalchemy.exc import IntegrityError
from sqlmodel import select

from ....tasks import collect_tasks
from ...db import async_session_maker
from ...db import AsyncSession
from ...db import DBSyncSession
from ...db import get_db
from ...db import get_sync_db
from ...models import SubtaskCreate
from ...models import Task
from ...models import TaskCreate
from ...models import TaskRead
from ...models import TaskUpdate
from ...security import current_active_user
from ...security import User

router = APIRouter()


async def upsert_task(
    task: Dict[str, Any],
) -> str:
    async with async_session_maker() as db:
        task_obj = TaskCreate(**task)
        try:
            task_orm = Task.from_orm(task_obj)
            db.add(task_orm)
            await db.commit()
            return "inserted"
        except IntegrityError:
            await db.rollback()
            stm = select(Task).where(Task.name == task_obj.name)
            res = await db.execute(stm)
            this_task = res.scalars().one()
            for key, value in task_obj.dict(exclude={"subtask_list"}).items():
                setattr(this_task, key, value)
            db.add(this_task)
            await db.commit()
            return "updated"


async def collect_tasks_headless():
    out = dict(inserted=0, updated=0)
    results = await asyncio.gather(
        *[upsert_task(task) for task in collect_tasks()]
    )
    results = sorted(results)
    out.update(dict(Counter(results)))
    return out


@router.post("/collect/", status_code=status.HTTP_201_CREATED)
async def collect_core_tasks(
    user: User = Depends(current_active_user),
):
    return await collect_tasks_headless()


@router.get("/", response_model=List[TaskRead])
async def get_list_task(
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    stm = select(Task)
    res = await db.execute(stm)
    task_list = res.scalars().unique().fetchall()
    await asyncio.gather(*[db.refresh(t) for t in task_list])
    return task_list


@router.get("/{task_id}", response_model=TaskRead)
def get_task(
    task_id: int,
    user: User = Depends(current_active_user),
    db_sync: DBSyncSession = Depends(get_sync_db),
):
    task = db_sync.get(Task, task_id)
    return task


@router.post("/", response_model=TaskRead, status_code=status.HTTP_201_CREATED)
async def create_task(
    task: TaskCreate,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):

    # Check that there is no task with the same user and name
    stm = select(Task).where(Task.name == task.name)
    res = await db.execute(stm)
    if res.scalars().all():
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Task name ({task.name}) already in use",
        )

    db_task = Task.from_orm(task)
    db.add(db_task)
    await db.commit()
    await db.refresh(db_task)
    return db_task


@router.patch("/{task_id}", response_model=TaskRead)
async def patch_task(
    task_id: int,
    task_update: TaskUpdate,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):

    # FIXME add user-owned tasks

    db_task = await db.get(Task, task_id)

    for key, value in task_update.dict(exclude_unset=True).items():
        if key == "name":
            setattr(db_task, key, value)
        elif key == "default_args":
            current_default_args = deepcopy(db_task._arguments)
            current_default_args.update(value)
            setattr(db_task, key, current_default_args)
        else:
            raise Exception("patch_task endpoint cannot set {key=}")

    await db.commit()
    await db.refresh(db_task)
    return db_task


@router.post(
    "/{parent_task_id}/subtask/",
    response_model=TaskRead,
    status_code=status.HTTP_201_CREATED,
)
async def add_subtask(
    parent_task_id: int,
    subtask: SubtaskCreate,
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
    db_sync: DBSyncSession = Depends(get_sync_db),
):
    parent_task = await db.get(Task, parent_task_id)
    await parent_task.add_subtask(
        db=db,
        **subtask.dict(exclude={"parent_task_id"}),
    )

    # Fetch parent task synchronously to lazily resolve children
    parent_task = db_sync.get(Task, parent_task_id)
    return parent_task



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original author(s):
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
from fastapi import APIRouter
from fastapi import Depends

from ...db import AsyncSession
from ...db import get_db
from ...security import current_active_user
from ...security import User

router = APIRouter()


@router.post("/")
async def create_workflow(
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    """
    Add new global workflow
    """
    raise NotImplementedError


@router.post("/{wofklow_id}")
async def add_task(
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    """
    Add global task to global workflow
    """
    raise NotImplementedError


@router.get("/")
async def get_list_workflow(
    user: User = Depends(current_active_user),
    db: AsyncSession = Depends(get_db),
):
    """
    List public workflows
    """
    raise NotImplementedError
